// TTGO_Digi v1.02 : Programme TTGO LoRa.
// Ecrit par Frank LOBA (www.f4hdb.fr).
// Date création : 31/03/2023.
// Dernière mise à jour : 02/04/2023.


/**** VERSIONS ****/
// 01/04/2023 v1.00 Le digipeater est toujours actif.
// 02/04/2023 v1.01 Ajout de la fonction d'activation ou non du digipeater, et de la commande d'émission d'une trame manuelle avec le bouton du milieu.
// 02/04/2023 v1.02 Modification des délais d'émission. On émet après avoir attendu.


/**** INCLUSION DES FICHIERS EN-TETE ****/
#include <SPI.h>
#include <LoRa.h>

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#include <HardwareSerial.h>
#include <TinyGPSPlus.h>

#include <math.h>


/**** DEFINITION DES CONSTANTES ****/
#define OLED_SDA 21
#define OLED_SCL 22 
#define OLED_RST 16

#define OLED_SCREEN_WIDTH   128
#define OLED_SCREEN_HEIGHT  64

#define OLED_ADRESS 0x3C


#define LORA_SCK  5
#define LORA_SS   18
#define LORA_MISO 19  
#define LORA_MOSI 27
#define LORA_RST  23
#define LORA_DIO0 26

#define LORA_FREQUENCE         433775000L
#define LORA_SPREADING_FACTOR  12
#define LORA_SIGNAL_BANDWIDTH  125000L
#define LORA_CODING_RATE_4     5
#define LORA_SYNC_WORD         0x12

#define LORA_START_BYTE_1 '<'
#define LORA_START_BYTE_2 0xFF
#define LORA_START_BYTE_3 0x01


#define GPS_SERIAL_NUM  1
#define GPS_RX_PIN      34
#define GPS_TX_PIN      12

// Période d'émission en minutes.
#define APRS_PERIODE_TX   10                           
// Distance parcourue en mètres pour émettre.
#define APRS_DISTANCE_TX  1000                        
// Indicatif de l'émetteur.
#define APRS_SOURCE       "F4XXX-12"                  
// Indicatif du destinataire avec le chemin.
#define APRS_DESTINATAIRE "APLG01,WIDE1-1,WIDE2-1"
// Overlay APRS.    
#define APRS_OVERLAY      '/'
// Symbole APRS.
#define APRS_SYMBOLE      '['
// Commentaire de la trame APRS.
#define APRS_COMMENTAIRE  "www.f4hdb.fr"


#define TNC2_SSID_CAR '-'
#define TNC2_DST_CAR  '>'
#define TNC2_VIA_CAR  ','
#define TNC2_MSG_CAR  ':'
#define TNC2_RPT_CAR  '*'


#define AX25_NAME_SIZE  6
#define AX25_SPACE      ' '
#define AX25_MAX_RPT    8


// Le boutton du module TTGO T-Beam est sur la broche 38.
#define TBEAM_BUTTON_PIN    38
// Délai d'attente en millisecondes pour permuter l'état d'activation, ou non, du digipeater.
#define TBEAM_BUTTON_DELAY  5000


/**** DECLARATION DES FONCTIONS GLOBALES ****/
String LatitudeAprs(RawDegrees Latitude);
String LongitudeAprs(RawDegrees Longitude);

String Digipeater(String TrameSrc);


/**** DEFINITION DES VARIABLES GLOBALES ****/
Adafruit_SSD1306 Display(OLED_SCREEN_WIDTH, OLED_SCREEN_HEIGHT, &Wire, OLED_RST);

// Trame APRS au format TNC2.
String TrameTNC2;

HardwareSerial GPSSerial(GPS_SERIAL_NUM);
TinyGPSPlus Gps;

// Ancien instant d'émission.
int InstantTx;

// Distance parcourue.
double Distance;  

// Latitude d'émission.
double LatitudeTx;

// Longitude d'émission.
double LongitudeTx;

// Trame du Digipeater.
String TrameDigi;

// Ancien instant d'appuis sur le bouton.
int InstantBtn;

// Etat d'activation ou non du digipeater.
bool Digi;

// Etat de l'émission d'une balise.
bool Beacon;


/**** DEFINITION DE LA FONCTION D'INITIALISATION ****/
void setup()
{
  // Initialisation de l'interface série (USB).
  Serial.begin(115200);

  // Initialisation de l'écran OLED.
  Wire.begin(OLED_SDA, OLED_SCL);
  
  if(!Display.begin(SSD1306_SWITCHCAPVCC, OLED_ADRESS, false, false))
  {
    // Ecriture d'un message d'erreur dans le port série en cas d'échec.
    Serial.println("Echec initialisation ecran");
    
    while(1);
  }

  // Effacement de l'écran, définition de la couleur et de la taille du texte.
  Display.clearDisplay();
  Display.setTextColor(WHITE);
  Display.setTextSize(1);

  // Affectation des broches pour LoRa.
  SPI.begin(LORA_SCK, LORA_MISO, LORA_MOSI, LORA_SS);
  LoRa.setPins(LORA_SS, LORA_RST, LORA_DIO0);


  // Défintion de la fréquence et démarrage du LoRa.
  if(!LoRa.begin(LORA_FREQUENCE))
  {
    // Ecriture d'un message d'erreur dans le port série en cas d'échec.
    Serial.println("Echec demarrage LoRa");

    // Affichage d'un message à l'écran en cas d'échec.
    Display.setCursor(0,0);
    Display.println("Echec demarrage LoRa");
    Display.display();
    
    while (1);
  }

  // Définition des paramètres système pour LoRa.
  LoRa.setSpreadingFactor(LORA_SPREADING_FACTOR);
  LoRa.setSignalBandwidth(LORA_SIGNAL_BANDWIDTH);
  LoRa.setCodingRate4(LORA_CODING_RATE_4);
  LoRa.enableCrc();
  LoRa.setSyncWord(LORA_SYNC_WORD);

  // Ecriture d'un message dans le port série en cas de succès.
  Serial.println("LoRa OK");

  // Affichage d'un message à l'écran en cas de succès.
  Display.setCursor(0,0);
  Display.println("LoRa OK");
  Display.display();

  // Initialisation du GPS.
  GPSSerial.begin(9600, SERIAL_8N1, GPS_RX_PIN, GPS_TX_PIN);

  // Initialisation de l'instant d'émission.
  InstantTx = millis();

  // Initialisation de la distance parcoure.
  Distance = 0.0;

  // Initialisation de la latitude d'émission.
  LatitudeTx = 0.0;

  // Initialisation de la longitude d'émission.
  LongitudeTx = 0.0;

  // Initialisation digipeater.
  Digi = true;

  // Intialisation de l'émission d'une balise.
  Beacon = false;

  // Initialisation du bouton.
  pinMode(TBEAM_BUTTON_PIN, INPUT);

  // Attente de 5 secondes.
  delay(5000);
}


/**** DEFINITION DE LA FONCTION D'EXECUTION ****/
void loop()
{
  // Détection de l'appuis sur le button du module TTGO T-BEAM.
  if(digitalRead(TBEAM_BUTTON_PIN) == LOW)
  {
    // Mémorisation de l'instant de l'appuis sur le bouton.
    InstantBtn = millis();

    // On attend tant que le bouton est appuyé.
    while((digitalRead(TBEAM_BUTTON_PIN) == LOW) && (millis() - InstantBtn) < TBEAM_BUTTON_DELAY);

    // Si l'on appuis longtemps, alors on permute l'activation du digipeater.
    if((millis() - InstantBtn) >= TBEAM_BUTTON_DELAY)
    {
      if(Digi == false)
      {
        Digi = true;

        // Effacement de l'écran.
        Display.clearDisplay();
    
        // Positionnement du curseur.
        Display.setCursor(0,0);
    
        // Affichage de "Digipeater ON" à l'écran.
        Display.println("Digipeater ON");
        Display.display();

        // Attente de 5 secondes.
        delay(5000);
      }
      else
      {
        Digi = false;

        // Effacement de l'écran.
        Display.clearDisplay();
    
        // Positionnement du curseur.
        Display.setCursor(0,0);
    
        // Affichage de "Digipeater OFF" à l'écran.
        Display.println("Digipeater OFF");
        Display.display();    

        // Attente de 5 secondes.
        delay(5000);
      }
    }
    // Sinon on active l'émission d'une balise.
    else Beacon = true;
  }

  
  // Détection de la réception d'un paquet.
  int PacketSize = LoRa.parsePacket();
  if (PacketSize)
  {
    // Effacement de l'écran.
    Display.clearDisplay();

    // Positionnement du curseur.
    Display.setCursor(0,0);

    // Affichage de "RX" à l'écran.
    Display.println("**** RX ****");

    // Initialisation de la trame TNC2.
    TrameTNC2 = "";

    // Détection de l'en-tête LoRa.
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_1)
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_2)
    if(LoRa.available() && LoRa.read() == LORA_START_BYTE_3)
    
    // Lecture du paquet au format TNC2.
    while(LoRa.available())
    TrameTNC2 += (char)LoRa.read();

    // Affichage de la trame au format TNC2 à l'écran.
    Display.println(TrameTNC2);
    Display.display();

    // Digipeater.
    TrameDigi = Digipeater(TrameTNC2);
    
    // Attente de 5 secondes.
    delay(5000);

    // Le décodage a réussi.
    if((TrameDigi.length() > 0) && Digi)
    {
      // Effacement de l'écran.
      Display.clearDisplay();
  
      // Positionnement du curseur.
      Display.setCursor(0,0);
  
      // Affichage de "RX" à l'écran.
      Display.println("**** DIGI ****");

      // Affichage de la trame au format TNC2 à l'écran.
      Display.println(TrameDigi);     
      Display.display();

      // Attente de 5 secondes.
      delay(5000);

      // Commencement de l'émission LoRa.
      LoRa.beginPacket();
          
      // Emission de l'en-tête LoRa.
      LoRa.write(LORA_START_BYTE_1);
      LoRa.write(LORA_START_BYTE_2);
      LoRa.write(LORA_START_BYTE_3);
          
      // Emission de la trame TNC2 en LoRa.
      LoRa.write((const uint8_t *)TrameDigi.c_str(), TrameDigi.length());
          
      // Fin de l'émission LoRa.
      LoRa.endPacket();
    }
  }
  // Si l'on ne reçoit pas de trames, on effectue le traçage.
  else
  {
    while (GPSSerial.available() > 0)
    Gps.encode(GPSSerial.read());

    if(Gps.location.isUpdated() || Gps.time.isUpdated())
    {
      // Calcul de la distance parcourue en mètres.
      // La circonférence de la Terre est de 40000 Km.
      // On utilise le théorème de Pythagore et la trigonométrie.
      Distance = 1000.0 * sqrt(pow((40000.0 / 360.0 * (Gps.location.lat() - LatitudeTx)), 2) + pow((40000.0 * cos(Gps.location.lat()) / 360.0 * (Gps.location.lng() - LongitudeTx)), 2));
      
      // Effacement de l'écran.
      Display.clearDisplay();
  
      // Positionnement du curseur.
      Display.setCursor(0,0);
  
      // Affichage de "GPS" à l'écran.
      Display.println("**** GPS ****");

      Display.print("Latitude : ");
      Display.println(LatitudeAprs(Gps.location.rawLat()));
      Display.print("Longitude : ");
      Display.println(LongitudeAprs(Gps.location.rawLng()));
      Display.print("Satellites: ");
      Display.println(Gps.satellites.value());
      Display.print("Altitude : ");
      Display.print(Gps.altitude.feet() / 3.2808);
      Display.println("m");
      Display.print("Time : ");
      Display.print(Gps.time.hour());
      Display.print(":");
      Display.print(Gps.time.minute());
      Display.print(":");
      Display.println(Gps.time.second());
      Display.print("Distance : ");
      Display.print(Distance);
      Display.println("m");
      Display.display();
    }
  
    if(((Distance > APRS_DISTANCE_TX) || ((millis() - InstantTx) > (APRS_PERIODE_TX * 60000)) || Beacon) && Gps.location.isValid()) 
    {
      // Mise à jour de l'instant d'émission.
      InstantTx = millis();
      
      // Mise à jour de la latitude et la longitude d'émission.
      LatitudeTx = Gps.location.lat();
      LongitudeTx = Gps.location.lng();

      // Constitution de la trame APRS.
      TrameTNC2 = APRS_SOURCE;
      TrameTNC2 += '>';
      TrameTNC2 += APRS_DESTINATAIRE;
      TrameTNC2 += ":=";
      TrameTNC2 += LatitudeAprs(Gps.location.rawLat());
      TrameTNC2 += APRS_OVERLAY;
      TrameTNC2 += LongitudeAprs(Gps.location.rawLng());
      TrameTNC2 += APRS_SYMBOLE;
      TrameTNC2 += APRS_COMMENTAIRE;
      
      // Effacement de l'écran.
      Display.clearDisplay();

      // Positionnement du curseur.
      Display.setCursor(0,0);

      // Affichage de "TX" à l'écran.
      Display.println("**** TX ****");

      // Affichage de la trame au format TNC2 à l'écran.
      Display.println(TrameTNC2);     
      Display.display();

      // Désactivation de l'émission d'une balise.
      Beacon = false;

      // Attente de 5 secondes.
      delay(5000);

      // Commencement de l'émission LoRa.
      LoRa.beginPacket();
            
      // Emission de l'en-tête LoRa.
      LoRa.write(LORA_START_BYTE_1);
      LoRa.write(LORA_START_BYTE_2);
      LoRa.write(LORA_START_BYTE_3);
            
      // Emission de la trame TNC2 en LoRa.
      LoRa.write((const uint8_t *)TrameTNC2.c_str(), TrameTNC2.length());
            
      // Fin de l'émission LoRa.
      LoRa.endPacket();
    }
  }

  // Attente de 20 milli secondes.
  delay(20);
}


/**** DEFINITION DE LA FONCTION DE CONVERSION DE LA LATITUDE AU FORMAT APRS ****/
String LatitudeAprs(RawDegrees Latitude)
{
  char Tampon[20];
  
  sprintf(Tampon, "%02d%05.2f", Latitude.deg, Latitude.billionths / 1000000000.0 * 60.0);
  
  String Chaine(Tampon);

  if(Latitude.negative) Chaine += 'S';
  else Chaine += 'N';
  
  return Chaine;
}



/**** DEFINITION DE LA FONCTION DE CONVERSION DE LA LONGITUDE AU FORMAT APRS ****/
String LongitudeAprs(RawDegrees Longitude)
{
  char Tampon[20];

  sprintf(Tampon, "%03d%05.2f", Longitude.deg, Longitude.billionths / 1000000000.0 * 60.0);

  String Chaine(Tampon);

  if(Longitude.negative) Chaine += 'W';
  else Chaine += 'E';
  
  return Chaine;
}


/**** DEFINITION DE LA FONCTION DE DIGIPEATER ****/
String Digipeater(String TrameSrc)
{
  int     IndexTNC2 = 0;
  int     IndexRpt = 0, NbRpt = 0;
  String  Src = "";
  String  Dst = "";
  String  Rpt[AX25_MAX_RPT];
  String  Msg = "";
  char    Ssid;
  bool    IsRpt;
  String  TrameDst = "";

  // Lecture de la trame TNC2 source.

  // Détection d'une trame TNC2 source valide.
  // Elle doit contenir le caractère TNC2_DST_CAR et le caractère TNC2_MSG_CAR.
  if(TrameSrc.indexOf(TNC2_DST_CAR) == -1 || TrameSrc.indexOf(TNC2_MSG_CAR) == -1)
  {
    TrameDst = "";
    return TrameDst;
  }

  // Lecture de l'indicatif de la station source dans la trame TNC2 source.
  for(IndexTNC2 = 0; IndexTNC2 < TrameSrc.indexOf(TNC2_DST_CAR); IndexTNC2 ++)
  Src += TrameSrc.charAt(IndexTNC2);

  // Lecture de l'indicatif de la station de destination dans la trame TNC2 source.
  IndexTNC2 = TrameSrc.indexOf(TNC2_DST_CAR) + 1;
  
  while(TrameSrc.charAt(IndexTNC2) != TNC2_MSG_CAR && TrameSrc.charAt(IndexTNC2) != TNC2_VIA_CAR)
  {
    Dst += TrameSrc.charAt(IndexTNC2);
    IndexTNC2 ++;
  }

  // Lecture des indicatifs des répéteurs dans la trame TNC2 source.
  IndexRpt = 0;

  while(TrameSrc.charAt(IndexTNC2) == TNC2_VIA_CAR && IndexRpt < AX25_MAX_RPT)
  {
    IndexTNC2 ++;
    Rpt[IndexRpt] = "";
    
    while(TrameSrc.charAt(IndexTNC2) != TNC2_VIA_CAR && TrameSrc.charAt(IndexTNC2) != TNC2_MSG_CAR)
    {
      Rpt[IndexRpt] += TrameSrc.charAt(IndexTNC2);
      IndexTNC2 ++;
    }

    IndexRpt ++;
  }

  // Ecriture du nombre de stations répétitrices comptées dans la trame source au format TNC2.
  NbRpt = IndexRpt;

  // Lecture du message.
  if(TrameSrc.charAt(IndexTNC2) == TNC2_MSG_CAR)
  {
    Msg = TrameSrc.substring(TrameSrc.indexOf(TNC2_MSG_CAR) + 1);
  }
  else
  {
    TrameDst = "";
    return TrameDst;
  }


  // Détection de la présence de stations répétitrices.
  if(NbRpt > 0)
  {
    // Ecriture de l'indicatif de la station source dans la trame TNC2 de destination.
    TrameDst = Src;
    TrameDst += TNC2_DST_CAR;
    
    // Ecriture de l'indicatif de la station source dans la trame TNC2 de destination.
    TrameDst += Dst;
    TrameDst += TNC2_VIA_CAR;

    // Initialisation de la répétition.
    IsRpt = false;

    // Pour chaque station répétitrice.
    for(IndexRpt = 0; IndexRpt < NbRpt; IndexRpt ++)
    {
      // Si le répéteur est au format WIDE, et que l'on ne doit pas déjà répéter, on peut envisager de répéter.
      if((Rpt[IndexRpt].indexOf("WIDE") == 0) && !IsRpt)
      {
        // S'il n'y a pas le caractère * à la fin du répéteur, on doit envisager de répéter.
        if(Rpt[IndexRpt].indexOf(TNC2_RPT_CAR) == -1)
        {
          // S'il y a le caractère - pour le SSID, on doit envisager de répéter.
          if(Rpt[IndexRpt].indexOf(TNC2_SSID_CAR) != -1)
          {
            // Ecriture de l'indicatif du module dans la trame TNC2 de destination.
            TrameDst += APRS_SOURCE;
            TrameDst += TNC2_RPT_CAR;
            TrameDst += TNC2_VIA_CAR;

            // Ecriture du répéteur dans la trame TNC2 de destination.
            TrameDst += Rpt[IndexRpt].substring(0, Rpt[IndexRpt].indexOf(TNC2_SSID_CAR));
            
            // Lecture du SSID de la station répétitrice au format TNC2.
            Ssid = (char) Rpt[IndexRpt].substring(Rpt[IndexRpt].indexOf(TNC2_SSID_CAR) + 1).toInt();
            
            // Vérification de la conformité du SSID.
            if(Ssid >= 16 || Ssid == 0)
            {
              TrameDst = "";
              return TrameDst;
            }

            // Décrémentation du SSID.
            Ssid --;

            // Si le SSID vaut 0, on ajoute le caractère * dans la trame TNC2 de destination.
            if(Ssid == 0) TrameDst += TNC2_RPT_CAR;
            // Sinon on ajoute le caractère -, puis le SSID, dans la trame TNC2 de destination.
            else
            {
              TrameDst += TNC2_SSID_CAR;
              TrameDst += Ssid;
            }

            // Ecriture du caractère virgule s'il reste des répéteurs.
            if(IndexRpt < (NbRpt - 1)) TrameDst += TNC2_VIA_CAR;
 
            // Validation de la répétition.
            IsRpt = true;
          }
          //S'il n'y a pas le caractère - pour le SSID, on passe au répéteur suivant.
          else
          {
            TrameDst += Rpt[IndexRpt];

            // Ecriture du caractère virgule s'il reste des répéteurs.
            if(IndexRpt < (NbRpt - 1)) TrameDst += TNC2_VIA_CAR;

          }
        }
        // S'il y a le caractère * à la fin du répéteur, on passe au répéteur suivant.
        else
        {
          TrameDst += Rpt[IndexRpt];

          // Ecriture du caractère virgule s'il reste des répéteurs.
          if(IndexRpt < (NbRpt - 1)) TrameDst += TNC2_VIA_CAR;

        }
      }
      // Si le répéteur n'est pas au format WIDE, on passe au répéteur suivant.
      else
      {
        TrameDst += Rpt[IndexRpt];
 
        // Ecriture du caractère virgule s'il reste des répéteurs.
        if(IndexRpt < (NbRpt - 1)) TrameDst += TNC2_VIA_CAR;
      }
    }
  }
  // S'il n'y a pas de stations répétitrices, on ne répète pas.
  else
  {
    TrameDst = "";
    return TrameDst;
  }

  
  // Si l'on doit répéter, on écrit le message dans la trame TNC2 de destination.
  if(IsRpt)
  {
    TrameDst += TNC2_MSG_CAR;
    TrameDst += Msg;
  }
  // Sinon on efface la trame TNC2 de destination.
  else TrameDst = "";
  
  // On retourne la trame TNC2 de destination.
  return TrameDst;
}
